import argparse
import numpy as np
import json
import requests
import urllib
from python.proto import prediction_pb2
from python.proto import prediction_pb2_grpc
from python.microservice import array_to_list_value
import grpc

def gen_continuous(range,n):
    if range[0] == "inf" and range[1] == "inf":
        return np.random.normal(size=n)
    if range[0] == "inf":
        return range[1]-np.random.lognormal(size=n)
    if range[1] == "inf":
        return range[0]+np.random.lognormal(size=n)
    return np.random.uniform(range[0],range[1],size=n)
        
def reconciliate_cont_type(feature,dtype):
    if dtype=="FLOAT":
        return feature
    if dtype=="INT":
        return (feature+0.5).astype(int).astype(float)

def gen_categorical(values,n):
    vals = np.random.randint(len(values),size=n)
    return np.array(values)[vals].astype(float)

def generate_batch(contract,n):
    feature_batches = []
    for feature_def in contract['features']:
        if feature_def["ftype"] == "continuous":
            if "range" in feature_def:
                range = feature_def["range"]
            else:
                range = ["inf","inf"]
            batch = gen_continuous(range,n)
            batch = reconciliate_cont_type(batch,feature_def["dtype"])
        elif feature_def["ftype"] == "categorical":
            batch = gen_categorical(feature_def["values"],n)
        feature_batches.append(batch[:,None])
    return np.concatenate(feature_batches,axis=1)

def gen_REST_request(batch,features,tensor=True):
    if tensor:
        datadef = {
            "names":features,
            "tensor":{
                    "shape":batch.shape,
                    "values":batch.ravel().tolist()
                    }
            }
    else:
        datadef = {
            "names":features,
            "ndarray":batch.tolist()
            }
        
    request = {
        "meta":{},
        "data":datadef
        }

    return request

def gen_GRPC_request(batch,features,tensor=True):
    if tensor:
        datadef = prediction_pb2.DefaultData(
            names = features,
            tensor = prediction_pb2.Tensor(
                shape = batch.shape,
                values = batch.ravel().tolist()
                )
            )
    else:
        datadef = prediction_pb2.DefaultData(
            names = features,
            ndarray = array_to_list_value(batch)
            )
    request = prediction_pb2.SeldonMessage(
        data = datadef
        )
    return request

def unfold_contract(contract):
    unfolded_contract = {}
    unfolded_contract["targets"] = []
    unfolded_contract["features"] = []

    for feature in contract["features"]:
        if feature.get("repeat") is not None:
            for i in range(feature.get("repeat")):
                new_feature = {}
                new_feature.update(feature)
                new_feature["name"] = feature["name"]+':'+str(i)
                del new_feature["repeat"]
                unfolded_contract["features"].append(new_feature)
        else:
            unfolded_contract["features"].append(feature)

    for target in contract["targets"]:
        if target.get("repeat") is not None:
            for i in range(target.get("repeat")):
                new_target = {}
                new_target.update(target)
                new_target["name"] = target["name"]+":"+str(i)
                del new_target["repeat"]
                unfolded_contract["targets"].append(new_target)
        else:
             unfolded_contract["targets"].append(target)   

    return unfolded_contract
                

def run(args):
    contract = json.load(open(args.contract,'r'))
    contract = unfold_contract(contract)
    feature_names = [feature["name"] for feature in contract["features"]]

    REST_url = "http://"+args.host+":"+str(args.port)+"/predict"

    for i in range(args.n_requests):
        batch = generate_batch(contract,args.batch_size)
        if args.prnt:
            print('-'*40)
            print("SENDING NEW REQUEST:")
        
        if not args.grpc:
            REST_request = gen_REST_request(batch,features=feature_names,tensor=args.tensor)
            if args.prnt:
                print(REST_request)
            
            response = requests.post(
                REST_url,
                data={"json":json.dumps(REST_request),"isDefault":True})
            jresp = response.json()

            if args.prnt:
                print("RECEIVED RESPONSE:")
                print(jresp)
                print()
        else:
            GRPC_request = gen_GRPC_request(batch,features=feature_names,tensor=args.tensor)
            if args.prnt:
                print(GRPC_request)

            channel = grpc.insecure_channel('{}:{}'.format(args.host,args.port))
            stub = prediction_pb2_grpc.ModelStub(channel)
            response = stub.Predict(GRPC_request)
            
            if args.prnt:
                print("RECEIVED RESPONSE:")
                print(response)
                print()
    

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("contract",type=str,help="File that contains the data contract")
    parser.add_argument("host",type=str)
    parser.add_argument("port",type=int)
    parser.add_argument("-b","--batch-size",type=int,default=1)
    parser.add_argument("-n","--n-requests",type=int,default=1)
    parser.add_argument("--grpc",action="store_true")
    parser.add_argument("-t","--tensor",action="store_true")
    parser.add_argument("-p","--prnt",action="store_true",help="Prints requests and responses")

    args = parser.parse_args()

    run(args)
